/************************************************************************/

// This file is original from file manager sarah

/************************************************************************/

//#include	<gnome.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<dirent.h>
#include	<fcntl.h>
#include	"file.h"

/************************************************************************/
//		private functions on define
/************************************************************************/

/************************************************************************/
// allow a new glist just with dir ( remove ./ ../ )
GList	*get_dirs( gchar const *path )
{
GList			*list = NULL;
DIR				*dir;
struct dirent	*file;

	if( !( dir = opendir( path ) ) ) return( NULL );	// unknow dir
	
	// read and check all files in dir
	while( ( file	= readdir( dir ) ) )
	{	// discard current and parent
		if ( !strcmp( file->d_name, "." ) || !strcmp( file->d_name, ".." ) ) continue;
		
		if( S_ISDIR( get_real_file_type( path, file->d_name ) ) )
			list = g_list_append( list, (gpointer)g_strdup( file->d_name ) );		
	}
	closedir( dir );
	
return( list_by_name( list ) );
}

/************************************************************************/
// same but return just readable dir
GList	*get_dirs_with_access( gchar const *path )
{
GList			*list = NULL;
DIR				*dir;
gchar			*tmp;
struct dirent	*file;

	if( !( dir = opendir( path ) ) ) return( NULL );	// unknow dir
	
	// read and check all files in dir
	while( ( file	= readdir( dir ) ) )
	{	// discard current and parent
		if ( !strcmp( file->d_name, "." ) || !strcmp( file->d_name, ".." ) )
			continue;
		
		if( !S_ISDIR( get_real_file_type( path, file->d_name ) ) )
			continue;
		
		tmp = g_strconcat( path, "/", file->d_name, NULL );
		
		if( !access( tmp, R_OK | X_OK ) )
			list = g_list_append( list, (gpointer)g_strdup( file->d_name ) );		

		g_free( tmp );	
	}
	closedir( dir );
	
return( list_by_name( list ) );
}

/************************************************************************/
// return FALSE if OK
 gboolean file_exist(gchar const *path, gchar const *name)
{
gchar			*f_path = g_strconcat( path, "/", name, NULL );
gint register	result;

	result = access( f_path, F_OK );
	g_free( f_path );
	
	return( result );
}

/************************************************************************/
// return FALSE if OK
gboolean path_exist( gchar const *path )
{
struct stat		st;

	if( stat( path, &st ) || access( path, R_OK | X_OK ) )
		return( TRUE );						// path not found

	return( !S_ISDIR( st.st_mode ) );		// exit but is it a dir ??
}

/************************************************************************/
// return type of file ( according to sys/stat.h ) or -1 on error
mode_t get_file_type( gchar const *path, gchar const *file )
{
struct stat		st;
gchar			*full_path = g_strconcat( path, "/", file, NULL );
mode_t			ret = -1;

	if( !lstat( full_path, &st ) )
		ret = st.st_mode & __S_IFMT;		// all right
	g_free( full_path );
	return( ret );
}

/************************************************************************/
// return type of file ( according to sys/stat.h ) or -1 on error ; fallow link
mode_t get_real_file_type( gchar const *path, gchar const *file )
{
struct stat		st;
gchar			*full_path = g_strconcat( path, "/", file, NULL );
mode_t			ret = -1;

	if( !stat( full_path, &st ) )
		ret = st.st_mode & __S_IFMT;		// all right
		
	else if( !lstat( full_path, &st ) )
		ret = st.st_mode & __S_IFMT;		// just an empty link

	g_free( full_path );
	return( ret );
}
