/************************************************************************/

// original case from file manager sarah

/************************************************************************/


#include	<gnome.h>
#include	<string.h>
#include	<sys/stat.h>
#include	<sys/wait.h>
#include	"file_op.h"
#include	"file.h"

/************************************************************************/

/************************************************************************/
GList	*list_by_name( GList *list )
{
gpointer	tmp;

	if( !list ) return( NULL );			// prevent from null ptr
	list = g_list_first( list );

	while( list->next )
	{	if( strcmp( (gchar *)list->data, (gchar *)list->next->data ) > 0 )
		{	tmp = list->data;				// greater ; must loop
			list->data = list->next->data;
			list->next->data = tmp;
			list = g_list_first( list );
		}
		else	// ok next cmp
			list = g_list_next( list );		
	}
	
return( g_list_first( list ) );
}

/************************************************************************/
// same as g_dirname, execpt that can return parent of a dir ( p ending with '/' )
// always return a slash ending string, except for no solution, return "."
gchar	*g_dirname_ext( gchar const *p )
{
gchar	*r = g_strdup( p );
gchar	*n;

	if( !strcmp( "/", p ) )
		return( g_strdup( "/" ) );

	// want parent of a dir ???
	if( r[ strlen( r ) - 1 ] == '/' )
		r[ strlen( r ) - 1 ] = '\0';
		
	if( !( n = strrchr( r, '/' ) ) )
	{	g_free( r );
		return( g_strdup( "." ) );
	}

	// remove multiple slash sequence
	while( *( n - sizeof( char ) ) == '/' )
		n--;
	
	n[1] = '\0';					// preserve the ending slash ( 0 is the slash )
	
	// r is g_new allocated, can change some char to '\0', g_free will free the full size...
	return( r );
}

/************************************************************************/
// return ( g_new ) the last file / dir of a path string, even if end with '/'
// p must be a valid path form, no "/"...
gchar *g_basename_ext( gchar const *p )
{
gchar	*t, *u;

	if( p[ strlen( p ) - 1 ] != '/' )
		return( g_strdup( g_basename( p ) ) );

	t = g_strdup( p );
	
	t[ strlen( p ) - 1 ] = 0;
	u = g_strdup( g_basename( t ) );
	g_free( t );
	
	return( u );
}

/************************************************************************/
// create a dynamic path from source to dest, must be both static
// because start with root path, there always a valid return value
gchar	*made_dynamic_path( gchar const *source, gchar const *path_dest )
{
gchar		*result, **s, **d;
gchar		*path_src = g_dirname_ext( source );
GString		*chaine;
guint		i=1, j;

	chaine = g_string_new( "./" );
	s = g_strsplit( path_src, "/", -1 );
	d = g_strsplit( path_dest, "/", -1 );

	// break on first string diff : found common part of both dir
	while( s[i] && d[i] )
	{	if( strcmp( s[i], d[i] ) )
			break;
		i++;
	}

	// now add "../" for each dir lesf on dest path of link : go to root
	j = i;
	while( d[i++] )
		g_string_append( chaine, "../" );

	// now link is on common dir, go to directory of dest file
	while( s[j] )
	{	g_string_append( chaine, s[j++] );
		g_string_append_c( chaine, '/' );
	}

	// end : add the file to link string
	g_string_append( chaine, ( result = g_basename_ext( source ) ) );
	g_free( result );

	result = g_strdup( chaine->str );
	g_string_free( chaine, TRUE );
	g_strfreev( s );
	g_strfreev( d );
	g_free( path_src );
	
return( result );
}

/************************************************************************/
// create a static path from dynamic data of a file in path
// NULL on error
gchar	*made_static_path( gchar const *path, gchar const *data )
{
gchar	*chaine;
gchar	*result;
guint	i = 0;

	if( path[0] != '/' ) return( NULL );				// no reference
		
	if( data[0] == '/' ) return( g_strdup( data ) );	// already static
		
	chaine = g_strdup( path );
	if( chaine[ strlen( chaine ) -1 ] == '/' )
		chaine[ strlen( chaine ) -1 ] = 0;
	
	while( TRUE )
	{	if( !strncmp( &data[i], "./", 2 ) )
			i += 2;										// skip
		else if( !strncmp( &data[i], "../", 3 ) )
		{	gchar	*tmp;
		
			if( !( tmp = g_dirname( chaine ) ) )		// pas to parent dir
			{	g_free( chaine );
				return( NULL );							// no more data : error
			}
			
			g_free( chaine );
			chaine = tmp;
			i += 3;
		}
		else break;										// found start of real data
	}
	
	if( !data[i] )
		return( chaine );								// just a ' to parent dir link '
	
	result = g_strconcat( chaine, "/", &data[i], NULL );
	g_free( chaine );
		
	return( result );		
}

/************************************************************************/
// test if  ' child ' is child of ' parent ' directory
// nota : full static path
gboolean file_is_child_of( gchar const *parent_path, gchar *child_path )
{
gchar		**p, **c;
gboolean	result = TRUE;
guint		i = 0;								// may pass "/" as arg, can't start at 1

	p = g_strsplit( parent_path, "/", -1 );
	c = g_strsplit( child_path, "/", -1 );

	// if no dif until end of dir of parent, child is really child
	while( p[i] )
	{	if( !strcmp( p[i], c[i] ) )				// test if dif in name until end
			if( c[++i] )
				continue;						// all right, test next

		result = FALSE;
		break;
	}
	
	g_strfreev( p );
	g_strfreev( c );

	return( result );
}

/************************************************************************/
// true if one path is child of other
gboolean one_is_child( gchar const *path1, gchar const *path2 )
{
gchar		**p, **c;
gboolean	result = TRUE;
guint		i = 0;								// may pass "/" as arg, can't start at 1

	p = g_strsplit( path1, "/", -1 );
	c = g_strsplit( path2, "/", -1 );

	// if no dif until end of one path there one child of other
	while( i++, p[i] && c[i] )
	{	if( !strcmp( p[i], c[i] ) )				// test if dif in name until end
		{	result = FALSE;
			break;
		}
	}
	
	g_strfreev( p );
	g_strfreev( c );

	return( result );
}

/************************************************************************/
// 1 if path 1 child of 2, return 2 if path2 is child of 1, 0 if none
guint who_is_child(gchar const *path1, gchar const *path2 )
{
gchar		**p, **c;
guint		result = 2;
guint		i = 0;								// may pass "/" as arg, can't start at 1

	p = g_strsplit( path1, "/", -1 );
	c = g_strsplit( path2, "/", -1 );

	// if no dif until end of dir of parent, child is really child
	while( p[i] )
	{	if( strcmp( p[i], c[i] ) )				// test if dif in name until end
		{	result = 0;
			break;
		}

		if( !c[++i] )							// test end of path 2
		{	result = 1;
			break;
		}
	}

	g_strfreev( p );
	g_strfreev( c );

	return( result );
}

