/************************************************************************/

//			Gnome Path
//			Rev : 0.2

//			this implemant a widget for path browsing
//			currently depending to many function of the main soft Sarah

//			Author : J. LEVREAU (C)
//			E-mail : j.a.l@free.fr

//			Statut : free software for private use
//			This not a gpl soft : you can use all you want / need without restrictions
//			FOR PRIVATE USE OR FREE SOFTWARE ONLY

/************************************************************************/
#include	"conf_header"

#include	<gnome.h>
#include	"gnome_path.h"
#include	"file.h"
#include	"file_op.h"

/************************************************************************/
//		structure definition
/************************************************************************/
enum {
	PATH_CHANGED,
	LAST_SIGNAL
};

/************************************************************************/
//	static functions definition
/************************************************INSIDE_GNOME_COMMON************************/
static void gnome_path_class_init( GnomePathClass *class );
static void gnome_path_init( GnomePath *w );

static void	try_parent_callback( GnomePath *w, gpointer d );
static void	try_child_callback( GnomePath *w, gpointer d );
static void	try_fix_callback( GnomePath *w, gpointer d );
static void	changed_callback( GnomePath *w, gpointer d );

static gboolean tab_callback( GnomePath *w, GdkEventKey *e, gpointer d );

static gboolean throw_parent( GnomePath *w );

static void	drop_received( GtkWidget *w, GdkDragContext *c, gint x, gint y,
    GtkSelectionData *s, guint i, guint t, gpointer data);

/************************************************************************/
//	static variables definition
/************************************************************************/
static gint				gnome_path_signals[ LAST_SIGNAL ] = { 0 };

/************************************************************************/
//	functions code
/************************************************************************/
guint gnome_path_get_type( void )
{
static guint type = 0;

	if ( !type )
	{	GtkTypeInfo info = { "GnomePath", sizeof( GnomePath ), sizeof( GnomePathClass ),
			( GtkClassInitFunc )gnome_path_class_init,
			( GtkObjectInitFunc )gnome_path_init,
			( GtkArgSetFunc )NULL, ( GtkArgGetFunc )NULL };

		type = gtk_type_unique( gtk_hbox_get_type(), &info );
	}

	return( type );
}

/************************************************************************/
static void gnome_path_init( GnomePath *w )
{
static GtkTargetEntry default_target[] = { { "STRING",0, GDK_TARGET_STRING } };

GdkPixmap		*p;
GdkBitmap		*b;
GtkTooltips		*t;

	w->cp = g_strdup( g_get_home_dir() );
	w->oldpath = g_strdup( g_get_home_dir() );

	t = GTK_TOOLTIPS( gtk_tooltips_new() );

	gnome_stock_pixmap_gdk( GNOME_STOCK_BUTTON_PREV, NULL, &p, &b );
	gtk_container_add( GTK_CONTAINER( w->parent = GTK_BUTTON( gtk_button_new() ) ),	gtk_pixmap_new( p, b ) );
	gtk_box_pack_start( GTK_BOX( w ), GTK_WIDGET( w->parent ), FALSE, FALSE, 0 );
	gtk_widget_show( GTK_WIDGET( w->parent ) );
	gtk_tooltips_set_tip( t, GTK_WIDGET( w->parent ), "Go to parent", "Go to parent" );

	gtk_box_pack_start( GTK_BOX( w ), GTK_WIDGET( w->e = GTK_ENTRY( gtk_entry_new() ) ), TRUE, TRUE, 0 );
	gtk_entry_set_text( w->e, w->oldpath );
	gtk_widget_show( GTK_WIDGET( w->e ) );
	gnome_widget_add_help( GTK_WIDGET( w->e ), "Using Gnome Path Widget" );

	gnome_stock_pixmap_gdk( GNOME_STOCK_BUTTON_APPLY, NULL, &p, &b );
	gtk_container_add( GTK_CONTAINER( w->status = GTK_BUTTON( gtk_button_new() ) ),	
		GTK_WIDGET( w->icon = GTK_PIXMAP( gtk_pixmap_new( p, b ) ) ) );
	gtk_box_pack_start( GTK_BOX( w ), GTK_WIDGET( w->status ), FALSE, FALSE, 0 );
	gtk_widget_show( GTK_WIDGET( w->status ) );
	gtk_tooltips_set_tip( t, GTK_WIDGET( w->status ), "Fix bad path", "Fix bad path" );

	gnome_stock_pixmap_gdk( GNOME_STOCK_BUTTON_NEXT, NULL, &p, &b );
	gtk_container_add( GTK_CONTAINER( w->child = GTK_BUTTON( gtk_button_new() ) ),	gtk_pixmap_new( p, b ) );
	gtk_box_pack_start( GTK_BOX( w ), GTK_WIDGET( w->child ), FALSE, FALSE, 0 );
	gtk_widget_show( GTK_WIDGET( w->child ) );
	gtk_tooltips_set_tip( t, GTK_WIDGET( w->child ), "Go to child dir or open popup", "Go to child dir or open popup" );

	gtk_signal_connect_object( GTK_OBJECT( w->parent ), "clicked", GTK_SIGNAL_FUNC( try_parent_callback ), GTK_OBJECT( w ) );
	gtk_signal_connect_object( GTK_OBJECT( w->child ), "clicked", GTK_SIGNAL_FUNC( try_child_callback ), GTK_OBJECT( w ) );
	gtk_signal_connect_object( GTK_OBJECT( w->status ), "clicked", GTK_SIGNAL_FUNC( try_fix_callback ), GTK_OBJECT( w ) );
	gtk_signal_connect_object( GTK_OBJECT( w->e ), "changed", GTK_SIGNAL_FUNC( changed_callback ), GTK_OBJECT( w ) );
	gtk_signal_connect_object( GTK_OBJECT( w->e ), "key_press_event", GTK_SIGNAL_FUNC( tab_callback ), GTK_OBJECT( w ) );
		
	// add dnd support
    gtk_drag_dest_set( GTK_WIDGET( w ), GTK_DEST_DEFAULT_ALL, default_target, 1, GDK_ACTION_COPY );
    gtk_signal_connect( GTK_OBJECT( w ), "drag_data_received", GTK_SIGNAL_FUNC( drop_received ), NULL );
}

/************************************************************************/
static void gnome_path_class_init( GnomePathClass *class )
{
GtkObjectClass *object_class = ( GtkObjectClass* )class;
  
	gnome_path_signals[ PATH_CHANGED ] = 
		gtk_signal_new( "path_changed", GTK_RUN_FIRST, object_class->type,
		 GTK_SIGNAL_OFFSET( GnomePathClass, path_changed ),
		 gtk_marshal_NONE__ENUM, GTK_TYPE_NONE, 1, GTK_TYPE_ENUM );

	gtk_object_class_add_signals( object_class, gnome_path_signals, LAST_SIGNAL );
	class->path_changed = NULL;
}

/************************************************************************/
// test recur all parent until found a valid dir, change and call event in case
// use w->cp and NOT entry do nothing on no match
static gboolean throw_parent( GnomePath *w )
{
gchar	*tmp;

	while( strcmp( w->cp, "/" ) )
	{	if( !path_exist( w->cp ) )
		{	gtk_entry_set_text( w->e, w->cp );
			return( TRUE );
		}

		tmp = g_dirname_ext( w->cp );
		g_free( w->cp );
		w->cp = tmp;
	}

	gtk_entry_set_text( w->e, w->cp );
	return( TRUE );
}

/************************************************************************/
GnomePath	*gnome_path_new( void )			// create with emply path
{	return( GNOME_PATH( gtk_type_new( gnome_path_get_type() ) ) ); }

/************************************************************************/
// change the current path, fix it if wrong and flag set, call event if ok / fixed
GnomePathStatusEnum	gnome_path_set_path( GnomePath *w, gchar const *path, gboolean fix )
{
	g_return_val_if_fail( w != NULL, GNOME_PATH_STATUS_ERROR );
	g_return_val_if_fail( GNOME_IS_PATH( w ), GNOME_PATH_STATUS_ERROR );

	if( !path ) return( GNOME_PATH_STATUS_ERROR );
	g_free( w->cp );

	// may comput a static path from dynamic request
	if( path[0] == '/' )
		w->cp = g_strdup( path );
	else
		w->cp = made_static_path( gtk_entry_get_text( w->e ), path );
	
	// if path ok, set, -> emit "changed" event -> change icon, emit "path_changed"
	if( !path_exist( w->cp ) )
	{	gtk_entry_set_text( w->e, w->cp );					// change entry ( and emit signal )
		return( GNOME_PATH_STATUS_OK );
	}
	
	if( !fix )												// exit now ???
		return( GNOME_PATH_STATUS_ERROR );

	if( throw_parent( w ) )
		return( GNOME_PATH_STATUS_FIXED );
	
	return( GNOME_PATH_STATUS_ERROR );
}
				
/************************************************************************/
// fix the current path to a valid path ( may no change so call event according to flag ),
// call event on fix ; return TRUE on change / event call
gboolean gnome_path_fix_path( GnomePath *w, gboolean always_event )
{
	g_return_val_if_fail( w != NULL, FALSE );

	// first test if current is already valid		
	if( !path_exist( gtk_entry_get_text( w->e ) ) )
	{	if( always_event )
			gtk_signal_emit( GTK_OBJECT( w ), 
				gnome_path_signals[ PATH_CHANGED ], GNOME_PATH_CHANGE_NO_CHANGE );
		return( TRUE );
	}		

	g_free( w->cp );
	w->cp = g_strdup( gtk_entry_get_text( w->e ) );
	return( throw_parent( w ) );
}

/************************************************************************/
// try to complete ( and not change to child... ) the current ( invalid ) path to a solution
// if solution is NULL, auto test ( may return a ( gchar *) / GList of multiple solution
// if already valid, may also call event ( always_event )
// return NULL on no error ( or nothing to do ), -1 if there NO solution...
GList	*gnome_path_complete_path_with( GnomePath *w, gchar const *solution, gboolean always_event )
{
GList	*l, *s, *r = NULL;
gchar	*c, *t;

	g_return_val_if_fail( w != NULL, (GList *)(-1) );

	// test if current path already valid
	if( !path_exist( gtk_entry_get_text( w->e ) ) )
	{	if( always_event )
			gtk_signal_emit( GTK_OBJECT( w ), 
				gnome_path_signals[ PATH_CHANGED ],	GNOME_PATH_CHANGE_NO_CHANGE );
		return( NULL );
	}		

	if( solution )
	{	// solution is really a completion of current path ???
		if( !strstr( solution, g_basename( gtk_entry_get_text( w->e ) ) ) )
			return( (GList *)(-1) );										// current and sol no match

		g_free( w->cp );													// ok make change
		gtk_entry_set_text( w->e, ( w->cp = made_static_path( 
			( c = g_dirname_ext( gtk_entry_get_text( w->e ) ) ), solution ) ) );
		g_free( c );
	
		return( NULL );	
	}
	
	// get the valid dir list 
	g_free( w->cp );
	if( !( s = l = get_dirs( w->cp = g_dirname_ext( gtk_entry_get_text( w->e ) ) ) ) )
		return( (GList *)(-1) );											// no list, current is wrong

	// create a new list with all that match the current setting
	c = g_basename( gtk_entry_get_text( w->e ) );	

	while( l )
	{	t = strstr( (gchar *)l->data, c );
		
		if( t == (gchar *)l->data )
			r = g_list_append( r, l->data );					// match, 'move' data to new list
		else
			g_free( l->data );
		
		l = g_list_next( l );
	}
	g_list_free( s );

	// if list empty nothing math
	if( !r ) return( (GList *)(-1) );
	
	// if just one match, can update it directly
	if( g_list_length( ( r = g_list_first( r ) ) ) == 1 )
	{	gchar	*tmp = made_static_path( w->cp, (gchar *)r->data );
	
		gtk_entry_set_text( w->e, tmp );
		g_free( w->cp );
		g_free( r->data );
		g_list_free( r );
		w->cp = tmp;

		// add an ending "/", usefull when completion with keyboard
		gtk_entry_append_text( w->e, "/" );

		return( NULL );	
	}
	
	// multiple list, return it
	return( g_list_first( r ) );
}

/************************************************************************/
// TRUE for ok
gboolean gnome_path_complete_fast_path_with( GnomePath *w, gchar const *solution, gboolean always_event )
{
GList	*s, *l;

	s = l = gnome_path_complete_path_with( w, solution, always_event );

	if( !l ) return( TRUE );								// already ok
	if( l == (GList *)(-1) ) return( FALSE );				// can't complete

	// get a list of solution, we don't when these, free directly
	while( l )						
	{	g_free( (gchar *)l->data );
		l = g_list_next( l );
	}
	
	g_list_free( l );
	return( TRUE );
}

/************************************************************************/
// change current entry to is parent ; in case of acces denied can fix it ( try until root dir )
// keep on current if there are no solution
GnomePathStatusEnum gnome_path_change_to_parent( GnomePath *w, gboolean fix )
{
	g_return_val_if_fail( w != NULL, GNOME_PATH_STATUS_ERROR );

	// can't get parent of root !!!!!!!
	if( !strcmp( gtk_entry_get_text( w->e ), "/" ) )
		return( GNOME_PATH_STATUS_OK );
		
	g_free( w->cp );

	if( !path_exist( ( w->cp = g_dirname_ext( gtk_entry_get_text( w->e ) ) ) ) )
	{	gtk_entry_set_text( w->e, w->cp );					// change entry ( and emit signal )

//		if( strcmp( w->cp, "/" ) )
		// add an ending "/", usefull when completion with keyboard
//			gtk_entry_append_text( w->e, "/" );

		return( GNOME_PATH_STATUS_OK );
	}
	
	if( !fix )
		return( GNOME_PATH_STATUS_ERROR );

	if( throw_parent( w ) )
	{
		if( strcmp( w->cp, "/" ) )
		// add an ending "/", usefull when completion with keyboard
			gtk_entry_append_text( w->e, "/" );

		return( GNOME_PATH_STATUS_OK );
	}
			
	return( GNOME_PATH_STATUS_ERROR );
}

/************************************************************************/
// to this child 
GnomePathStatusEnum gnome_path_change_to_child( GnomePath *w, gchar const *child )
{
	g_return_val_if_fail( w != NULL, GNOME_PATH_STATUS_ERROR );
	g_free( w->cp );

	if( !path_exist( ( w->cp = made_static_path( gtk_entry_get_text( w->e ), child ) ) ) )
	{	gtk_entry_set_text( w->e, w->cp );					// change entry ( and emit signal )

		// add an ending "/", usefull when completion with keyboard
		gtk_entry_append_text( w->e, "/" );

		return( GNOME_PATH_STATUS_OK );
	}
	
	return( GNOME_PATH_STATUS_ERROR );
}

/************************************************************************/
// return the early valid path loaded
gchar *gnome_path_get_path( GnomePath *w )
{	
	g_return_val_if_fail( w != NULL, NULL );
	return( g_strdup( w->oldpath ) );
}

/************************************************************************/
// return the early valid path loaded, so as above but return const :
// don't free it and use it for quick function ( that don't risk to call gnome_path )
gchar const *gnome_path_get_path_fast( GnomePath *w )
{	
	g_return_val_if_fail( w != NULL, NULL );
	return( w->oldpath );
}

/************************************************************************/
// return the curreent content of entry widget, may wrong
gchar *gnome_path_get_widget_path( GnomePath *w )
{	
	g_return_val_if_fail( w != NULL, NULL );
	return( g_strdup( gtk_entry_get_text( w->e ) ) );
}

/************************************************************************/
//	private part : static function - callback function
/************************************************************************/
static void	try_parent_callback( GnomePath *w, gpointer d )
{	gnome_path_change_to_parent( w, FALSE ); }

/************************************************************************/
static void	try_child_callback( GnomePath *w, gpointer d )
{
GList		*l, *s;
GnomeUIInfo	*menu;
GtkWidget	*menu_wid;
guint		i;
gboolean	st;

	// this ...complete_path... return NULL if current path is valid
	// in case of an alone valid solution, ...complete_path... change to it and return NULL also
	st = path_exist( gtk_entry_get_text( w->e ) );

	// first, try to complete the current setting
	s = l = gnome_path_complete_path_with( w, NULL, FALSE );

	// now if the path change for inv to valid, must exit
	if( st && !path_exist( gtk_entry_get_text( w->e ) ) )
		return;
		
	// test if no match
	if( l == (GList *)(-1) )
	{	GnomeDialog	*d = GNOME_DIALOG( gnome_warning_dialog_parented( _( "No match." ),
			GTK_WINDOW( gtk_widget_get_toplevel( GTK_WIDGET( w ) ) ) ) );
	
		gnome_dialog_set_close( d, TRUE );
		gtk_widget_show( GTK_WIDGET( d ) );
		return;													// previous error, abort
	}
	
	// test if already valid : load child list and try 
	if( !l )
	{	s = l = get_dirs_with_access( gtk_entry_get_text( w->e ) );

		if( !l )
		{	GnomeDialog	*d = GNOME_DIALOG( gnome_ok_dialog_parented( _( "End of Path" ),
				GTK_WINDOW( gtk_widget_get_toplevel( GTK_WIDGET( w ) ) ) ) );
	
			gnome_dialog_set_close( d, TRUE );
			gtk_widget_show( GTK_WIDGET( d ) );
			return;
		}
	
		// if just 1 entry, can update directly
		if( g_list_length( l ) == 1 )
		{	g_free( w->cp );
			gtk_entry_set_text( w->e,
				( w->cp = made_static_path( gtk_entry_get_text( w->e ), (gchar *)l->data ) ) );
			g_free( l->data );
			g_list_free( l );
			return;
		}
		
		// multiple entry popup
		menu = g_new0( GnomeUIInfo, g_list_length( g_list_first( l ) ) + 1 );

		for( i=0, l = g_list_first( l ) ; l ; i++, l = g_list_next( l ) )
		{	menu[i].type = GNOME_APP_UI_ITEM;
			menu[i].label = (gchar *)l->data;
		}
		g_list_free( s );

		menu[i].type = GNOME_APP_UI_ENDOFINFO;
		menu_wid = gnome_popup_menu_new( menu );

		if( ( i = gnome_popup_menu_do_popup_modal( menu_wid, NULL, NULL, NULL, NULL ) ) != -1 )
		{	if( w->cp ) g_free( w->cp );
			gtk_entry_set_text( w->e,
				( w->cp = made_static_path( gtk_entry_get_text( w->e ), menu[i].label ) ) );

			// add an ending "/", usefull when completion with keyboard
			gtk_entry_append_text( w->e, "/" );
		}

		gtk_widget_destroy( menu_wid );		
		return;		
	}
	
	// open a list for user selection, create a popup menu
		menu = g_new0( GnomeUIInfo, g_list_length( g_list_first( l ) ) + 1 );

		for( i=0, l = g_list_first( l ) ; l ; i++, l = g_list_next( l ) )
		{	menu[i].type = GNOME_APP_UI_ITEM;
			menu[i].label = (gchar *)l->data;
		}
		g_list_free( s );

		menu[i].type = GNOME_APP_UI_ENDOFINFO;
		menu_wid = gnome_popup_menu_new( menu );

		if( ( i = gnome_popup_menu_do_popup_modal( menu_wid, NULL, NULL, NULL, NULL ) ) != -1 )
		{	gchar	*c;
			
			g_free( w->cp );
			gtk_entry_set_text( w->e, ( w->cp = made_static_path( 
				( c = g_dirname( gtk_entry_get_text( w->e ) ) ), menu[i].label ) ) );
			g_free( c );

			// add an ending "/", usefull when completion with keyboard
			gtk_entry_append_text( w->e, "/" );
		}

		gtk_widget_destroy( menu_wid );		
}

/************************************************************************/
static void	try_fix_callback( GnomePath *w, gpointer d )
{	gnome_path_fix_path( w, FALSE ); }

/************************************************************************/
static void	changed_callback( GnomePath *w, gpointer d )
{
GdkPixmap	*p;
GdkBitmap	*b;
GnomePathStatusEnum	s = GNOME_PATH_STATUS_ERROR;

	// if path is on error abort ( may is 'user write' )
	if( path_exist( gtk_entry_get_text( w->e ) ) )
	{	gnome_stock_pixmap_gdk( GNOME_STOCK_PIXMAP_STOP, NULL, &p, &b );
		gtk_pixmap_set( w->icon, p, b );
		return;
	}

	// test where this path respect to old valid path
	switch( who_is_child( w->oldpath, gtk_entry_get_text( w->e ) ) ) {
		case 0 :
			s = GNOME_PATH_CHANGE_TO_NEW;
			break;
		case 1 :
			s = GNOME_PATH_CHANGE_TO_PARENT;
			break;
		case 2 :	
			s = GNOME_PATH_CHANGE_TO_CHILD;
			break;
	}

	gnome_stock_pixmap_gdk( GNOME_STOCK_BUTTON_APPLY, NULL, &p, &b );
	gtk_pixmap_set( w->icon, p, b );

	g_free( w->oldpath );
	w->oldpath = g_strdup( gtk_entry_get_text( w->e ) );

	// all path must end with "/"
	if( w->oldpath[ strlen( w->oldpath ) - 1 ] != '/' )
	{	gchar	*tmp = w->oldpath;
	
		w->oldpath = g_strconcat( tmp, "/", NULL );
		g_free( tmp );
	}

	// emit callback
	gtk_signal_emit( GTK_OBJECT( w ), gnome_path_signals[ PATH_CHANGED ], s );
}

/************************************************************************/
static gboolean tab_callback( GnomePath *w, GdkEventKey *e, gpointer d )
{
	// test if must complete the current set
	if( ( e->keyval ==  GDK_Tab ) || ( e->keyval == GDK_Return ) )
	{	// is the same work as for child request !
		try_child_callback( w, d );	
		
		gtk_signal_emit_stop_by_name( GTK_OBJECT( w->e ), "key_press_event" );
		return( TRUE );							// stop the current event
	}

	// other key, 
	return( FALSE );
}

/************************************************************************/
// function called when a dnd is received into one of gnome_path widget
static void	drop_received( GtkWidget *w, GdkDragContext *c, gint x, gint y,
    GtkSelectionData *s, guint i, guint t, gpointer data )
{
	if( s->length < 1 ) return;									// one case of end		

	// can receive a file string, so must get the dirname first
	switch( get_real_file_type( "", s->data ) )
	{	case __S_IFLNK :
		case __S_IFREG :										// try dir name
		{	gchar	*path = g_dirname( s->data );

			if( gnome_path_set_path( GNOME_PATH( w ), path, FALSE ) != GNOME_PATH_STATUS_ERROR )
			{	g_free( path );		
				return;
			}

			g_free( path );
			break;												// direct to error popup message
		}
		case __S_IFDIR :
			if( gnome_path_set_path( GNOME_PATH( w ), s->data, FALSE ) != GNOME_PATH_STATUS_ERROR )
				return;											// change succesfull
		default :												// direct to error
	}
	
	// unknow path, popup dialog else already set and exit
	{	gchar		*str = g_strconcat( _( "Path unknow or denied :\n" ), s->data, NULL );
		GnomeDialog	*d = GNOME_DIALOG( gnome_error_dialog_parented( str,
			GTK_WINDOW( gtk_widget_get_toplevel( GTK_WIDGET( w ) ) ) ) );
	
		gnome_dialog_set_close( d, TRUE );
		gtk_widget_show( GTK_WIDGET( d ) );
		g_free( str );
	}
}

